#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <coral/io.h>
#include <coral/compat/unistd.h>
#include <coral/string.h>
#include <coral/memory.h>
#include <coral/inline.h>



enum {
	MMP_SOURCE,
	MMP_DEFFILE,
	MMP_RESOURCE,
	MMP_LIBRARY,
	MMP_AIF,
	MMP_EMIT,
	MMP_OPTION,
};



struct _mmp_tag {
	const char* tag;
	unsigned code;
} mmp_tag[] = {
	{ "SOURCE", MMP_SOURCE },
	{ "DEFFILE", MMP_DEFFILE },
	{ "RESOURCE", MMP_RESOURCE },
	{ "LIBRARY", MMP_LIBRARY },
	{ "AIF", MMP_AIF },
	{ "EMIT", MMP_EMIT },
	{ "OPTION", MMP_OPTION },
};



//const char pathfile[] = "mmppath.txt";
static char buf[256];



/* get_tag_code */

int
get_tag_code( const char* tag )
{
	int i;
	for( i=0; i<OP_ELEM_COUNT(mmp_tag); i++ ) {
		if( strcmpi(mmp_tag[i].tag,tag)==0 )
			return mmp_tag[i].code;
	}
	return -1;
}


static const char delimiter[] = "= ";
char token[256], tmpbuf[256];
struct _IOMEMORY iom;
struct _IOTEXT iot;
struct _MEMEXP mexp;

/* read_param */

static int
read_param( void )
{
	
	if( iotextRead(&iot,delimiter,&mexp,0,
		IOTEXT_FLAG_TRACE_DELIMITER)==IO_END ) {
		if( !token[0] ) return IO_END;
	}
	else {
		if( iot.lastchar=='=' ) return IO_CONTINUE;
	}
	return IO_ABNORMAL;
}

/* write_header */

FILE*
write_header( const char* makefile, const char* target_name,
	const char* option, const char* uid )
{
	FILE* fp = fopen( makefile, "w" );
	mexp.ctx = token;
	mexp.size = -(int)sizeof(token);

	memset( &iot, 0, sizeof(struct _IOTEXT) );
	ioInitializeMemoryBuffer
		( &iot.iob, &iom, (_integer1 __F*)option, strlen(option), 0 );

	iotextRead( &iot, NULL, &mexp, 0, 0 );

	if( fp ) {
		fprintf( fp,
			"TARGET  %s.%s\n"
			"TARGETTYPE  %s\n",
			target_name, token, token );
		if( strcmp(token,"dll")==0 ) {
			fprintf( fp, "UID  0x1000008d %s\n", uid );
		}
		else {
			fprintf( fp, "UID  0x100039CE %s\n"
				"TARGETPATH  \\system\\apps\\%s\n"
				, uid, target_name );
			while( iotextRead(&iot,delimiter,&mexp,0,
				IOTEXT_FLAG_TRACE_DELIMITER)!=IO_END ) {
				if( iot.lastchar=='=' ) {
					for(;;) {
						int result;
						switch( strFind("stack\0heap\0",token,0,
								FLAG_CASE_INSENSITIVE) ) {
							case 0:
							switch( read_param() ) {
								case IO_CONTINUE: continue;
								case IO_ABNORMAL:
								fprintf( fp, "EPOCSTACKSIZE %s\n", token );
							}
							break;
							case 1:
							switch( read_param() ) {
								case IO_END: break;
								case IO_CONTINUE: continue;
								default:
								strcpy( tmpbuf, token );
								result = read_param();
								fprintf( fp, "EPOCHEAPSIZE %s %s\n", tmpbuf,
									result==IO_ABNORMAL? token : tmpbuf );
								if( result==IO_CONTINUE ) continue;
							}
							break;
						}
						break;
					}
				}
			}
		}
	}
	return fp;
}



/* get_relative_path */

char*
alloc_relative_path( const char* filename )
{
	getcwd(buf,sizeof(buf)); /* MK_DEVEL */
	strAppendDirectorySeparator( buf );
	strcat( buf, "etc" );
	strAppendDirectorySeparator( buf );
	strcat( buf, getenv("MK_MAKEFILE") );
	strAppendDirectorySeparator( buf );
	return strGetRelativePath( buf, (char*)filename, STR_FLAG_ALLOCATE );
}



/* add_source */

int
add_source( FILE* fp, const char* tag, const char* filename )
{
	char * rel = alloc_relative_path( filename );
	if( rel ) {
//		FILE* pathfp;
//		char oldpath[256];
		filename = strGetFileName(rel);
		strGetPathName( rel );
		fprintf( fp, "SOURCEPATH %s\n", rel );
		fprintf( fp, "%s %s\n", tag, filename );
		memFree( rel );
	}
	return 0;
}



/* add_library */

int
add_library( FILE* fp, const char* const* filename )
{
	const char* fn;
	while( (fn=*filename++)!=NULL ) {
		fprintf( fp, "LIBRARY %s\n", fn );
	}
	return 0;
}



/* emit_string */

int
emit_string( FILE* fp, const char* const* tag )
{
	const char* t, * oldt=NULL;
	while( (t=*tag++)!=NULL ) {
		if( oldt ) fprintf( fp, " " );
		fprintf( fp, "%s", t );
		oldt = t;
	}
	fprintf( fp, "\n" );
	return 0;
}



/* add_aif */

int
add_aif( FILE* fp, const char* const * tag )
{
#if 1
	char* aifpath;
	fprintf( fp, "AIF %s", strGetFileName( *tag++ ) );
	aifpath = strCopy( NULL, *tag++, 0, strGetDirectorySeparatorSize() );
	if( aifpath ) {
		char* rel;
		strAppendDirectorySeparator( aifpath );
		rel = alloc_relative_path( aifpath );
		memFree( aifpath );
		if( rel ) {
			/* XXX define strRemoveDirectorySeparator */
			fprintf( fp, " %s ", rel );
			memFree( rel );
			return emit_string( fp, tag );
		}
	}
	return FAIL_MEMORY;
#else
	/* changed to generate all residuals into make directory */
	fprintf( fp, "AIF %s .. ", strGetFileName( *++tag ) );
	return emit_string( fp, tag+1 );
#endif
}


/* parse_option */

static char make_path[256];

const char*
parse_option( const char* argv[], FILE* fp )
{
	const char* o;
	_unsigned4 result;
	struct _STROPT opt;
	opt.getarg = stroptArgv;
	opt.flag = STROPT_FLAG_DEFAULT_INDICATOR|
		STROPT_FLAG_DEFAULT_DELIMITER|STROPT_FLAG_DOUBLE_INDICATOR;
	opt.context = (void*)argv;
	opt.append = NULL;
	opt.nextarg = NULL;

	while( (result=stroptGetOption
						( &opt, NULL, NULL, NULL, 0 ))!=STROPT_TERMINATED ) {
		if( result==STROPT_TERMINATED ) break;
		switch( result ) {

			case 'D':
			o = stroptGetString(&opt);
			if( fp ) {
				fprintf( fp, "MACRO %s\n", o );
			}
			break;

			case 'I':
			o = stroptGetString(&opt);
			if( fp ) {
				char buf[256];
				strcpy( buf, o );
				fprintf( fp, "USERINCLUDE %s\n",
					strGetRelativePath(make_path,buf,0) );
			}
			break;

			case STROPT_RESERVED:
			return opt.arg; 

			default:
			return NULL;
		}
	}
	return NULL;
}



/* main */

#if 0
int
main( int argc, const char* argv[] )
{
	const char* result = parse_option( argv+6 );
	puts( result? result : "null" );
	return 0;
}
#else
#ifdef __TURBOC__
#	pragma argsused
#else
#ifdef _MSC_VER
#	pragma warning( disable:_MSC_WARN_UNREFERENCED )
#endif
#endif
int
main( int argc, const char* argv[] )
{
	const char* filename = argv[1];
	int code = get_tag_code(argv[2]);
	FILE* fp = code!=MMP_SOURCE || access(filename,0)==0 ?
		fopen( filename, "a" ) :
		write_header( filename, argv[3], argv[4], argv[5] );
	if( fp ) {

		switch( code ) {

			case MMP_SOURCE:
			case MMP_RESOURCE:
			case MMP_DEFFILE:
			filename = parse_option( argv+5, NULL );
			if( filename==NULL || add_source( fp, argv[2], filename )<0 )
				printf( "Failed to add entry %s.\n", argv[6] );
			break;

			case MMP_OPTION:
			strcpy( make_path, getenv("MK_DEVEL") );
			strAppendDirectorySeparator(make_path);
			strcat( make_path, "make" );
			strAppendDirectorySeparator(make_path);
			strcat( make_path, "project" );
			strAppendDirectorySeparator(make_path);
			parse_option( argv+2, fp );
			break;

			case MMP_LIBRARY:
			if( add_library( fp, argv+3 )<0 )
				puts( "Failed to add library." );
			break;

			case MMP_EMIT:
			/* assume emission mode */
			if( emit_string( fp, argv+3 )<0 )
				puts( "Failed to emit strings." );
			break;

			case MMP_AIF:
			if( add_aif( fp, argv+6 ) <0 )
				puts( "Failed to add AIF." );
			break;

			default:
			printf( "Unknown tag %s.\n", argv[2] );
		}
		fclose( fp );
	}
	else {
		printf( "Failed to open the output file '%s'.", filename );
	}
	return 0;
}
#endif


/* EOF */

